@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2007-2013 Tristan Heaven <tristanheaven@gmail.com>
+ *   Copyright (C) 2007-2016 Tristan Heaven <tristan@tristanheaven.net>
  *
  *   This file is part of GtkHash.
  *
@@ -14,7 +14,7 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with GtkHash. If not, see <http://www.gnu.org/licenses/>.
+ *   along with GtkHash. If not, see <https://gnu.org/licenses/gpl-2.0.txt>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -24,6 +24,7 @@
 #include <stdlib.h>
 #include <stdbool.h>
 #include <gtk/gtk.h>
+#include <gdk/gdk.h>
 
 #include "callbacks.h"
 #include "main.h"
@@ -31,6 +32,8 @@
 #include "hash.h"
 #include "prefs.h"
 #include "list.h"
+#include "check.h"
+#include "uri-digest.h"
 #include "hash/hash-string.h"
 
 static bool on_window_delete_event(void)
@@ -41,93 +44,70 @@ static bool on_window_delete_event(void)
 	return true;
 }
 
+static void on_menuitem_open_activate(void)
+{
+	GtkFileChooser *chooser = GTK_FILE_CHOOSER(
+		gtk_file_chooser_dialog_new(_("Check Digests"), gui.window,
+			GTK_FILE_CHOOSER_ACTION_OPEN,
+			_("_Cancel"), GTK_RESPONSE_CANCEL,
+			_("_Open"), GTK_RESPONSE_ACCEPT,
+			NULL));
+	gtk_file_chooser_set_select_multiple(chooser, true);
+	gtk_file_chooser_set_local_only(chooser, false);
+
+	GtkFileFilter *filter = NULL;
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter,
+		_("Digest/Checksum Files (*.sha1, *.md5, *.sfv, ...)"));
+	check_file_add_filters(filter);
+	gtk_file_chooser_add_filter(chooser, filter);
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter, _("All Files"));
+	gtk_file_filter_add_pattern(filter, "*");
+	gtk_file_chooser_add_filter(chooser, filter);
+
+	if (gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
+		GSList *files = gtk_file_chooser_get_files(chooser);
+		GSList *ud_list = NULL;
+
+		for (GSList *p = files; p; p = p->next)
+			if (p->data)
+				ud_list = check_file_load(ud_list, p->data);
+
+		if (ud_list) {
+			if (gui.view == GUI_VIEW_FILE_LIST)
+				gui_add_ud_list(ud_list, GUI_VIEW_FILE_LIST);
+			else
+				gui_add_ud_list(ud_list, GUI_VIEW_INVALID);
+
+			gui_update();
+
+			uri_digest_list_free_full(ud_list);
+		}
+
+		g_slist_free_full(files, g_object_unref);
+	}
+
+	gtk_widget_destroy(GTK_WIDGET(chooser));
+}
+
 static void on_menuitem_save_as_activate(void)
 {
 	GtkFileChooser *chooser = GTK_FILE_CHOOSER(
 		gtk_file_chooser_dialog_new(_("Save Digests"), gui.window,
 			GTK_FILE_CHOOSER_ACTION_SAVE,
-			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-			GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
+			_("_Cancel"), GTK_RESPONSE_CANCEL,
+			_("_Save"), GTK_RESPONSE_ACCEPT,
 			NULL));
+	gtk_file_chooser_set_select_multiple(chooser, false);
+	gtk_file_chooser_set_local_only(chooser, true); // TODO
 	gtk_file_chooser_set_do_overwrite_confirmation(chooser, true);
 
-
 	if (gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
 		char *filename = gtk_file_chooser_get_filename(chooser);
-		GString *string = g_string_sized_new(1024);
-
-		for (int i = 0; i < HASH_FUNCS_N; i++) {
-			if (!hash.funcs[i].enabled)
-				continue;
-
-			switch (gui_get_view()) {
-				case GUI_VIEW_FILE: {
-					const bool hmac_active = gtk_toggle_button_get_active(
-						gui.togglebutton_hmac_file);
-					const char *digest = gtk_entry_get_text(
-						gui.hash_widgets[i].entry_file);
-					if (digest && *digest) {
-						g_string_append_printf(string,
-							(hmac_active && (hash.funcs[i].block_size > 0)) ?
-							"# HMAC-%s\n" : "# %s\n", hash.funcs[i].name);
-					} else
-						continue;
-					char *path = gtk_file_chooser_get_filename(
-						GTK_FILE_CHOOSER(gui.filechooserbutton));
-					char *basename = g_path_get_basename(path);
-					g_free(path);
-					g_string_append_printf(string, "%s  %s\n",
-					gtk_entry_get_text(gui.hash_widgets[i].entry_file),
-						basename);
-					g_free(basename);
-					break;
-				}
-				case GUI_VIEW_TEXT: {
-					const bool hmac_active = gtk_toggle_button_get_active(
-						gui.togglebutton_hmac_text);
-					g_string_append_printf(string,
-						(hmac_active && (hash.funcs[i].block_size > 0)) ?
-						"# HMAC-%s\n" : "# %s\n", hash.funcs[i].name);
-					g_string_append_printf(string, "%s  \"%s\"\n",
-						gtk_entry_get_text(gui.hash_widgets[i].entry_text),
-						gtk_entry_get_text(gui.entry_text));
-					break;
-				}
-				case GUI_VIEW_FILE_LIST: {
-					int prev = -1;
-					for (unsigned int row = 0; row < list_count_rows(); row++)
-					{
-						char *digest = list_get_digest(row, i);
-						if (digest && *digest) {
-							if (i != prev)
-								g_string_append_printf(string, "# %s\n",
-									hash.funcs[i].name);
-							prev = i;
-						} else {
-							if (digest)
-								g_free(digest);
-							prev = i;
-							continue;
-						}
-						char *uri = list_get_uri(row);
-						char *basename = g_filename_display_basename(uri);
-						g_string_append_printf(string, "%s  %s\n",
-							digest, basename);
-						g_free(basename);
-						g_free(uri);
-						g_free(digest);
-					}
-					break;
-				}
-				default:
-					g_assert_not_reached();
-			}
-		}
-
-		char *data = g_string_free(string, false);
-		g_file_set_contents(filename, data, -1, NULL);
-
-		g_free(data);
+		check_file_save(filename);
 		g_free(filename);
 	}
 
@@ -210,8 +190,35 @@ static void on_menuitem_prefs_activate(void)
 	gtk_widget_show(GTK_WIDGET(gui.dialog));
 }
 
+static void on_radiomenuitem_toggled(void)
+{
+	enum gui_view_e view = GUI_VIEW_INVALID;
+
+	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
+		gui.radiomenuitem_file)))
+	{
+		view = GUI_VIEW_FILE;
+	} else if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
+		gui.radiomenuitem_text)))
+	{
+		view = GUI_VIEW_TEXT;
+	} else if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
+		gui.radiomenuitem_file_list)))
+	{
+		view = GUI_VIEW_FILE_LIST;
+	}
+
+	g_assert(GUI_VIEW_IS_VALID(view));
+
+	if (gui.view != view) {
+		gui.view = view;
+		gui_update();
+	}
+}
+
 static void on_menuitem_about_activate(void)
 {
+#if (GTK_MAJOR_VERSION < 3)
 	static const char * const license = {
 		"This program is free software: you can redistribute it and/or modify\n"
 		"it under the terms of the GNU General Public License as published by\n"
@@ -222,36 +229,52 @@ static void on_menuitem_about_activate(void)
 		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
 		"GNU General Public License for more details.\n\n"
 		"You should have received a copy of the GNU General Public License along\n"
-		"with this program; if not, see <http://www.gnu.org/licenses/>.\n"
+		"with this program; if not, see <https://gnu.org/licenses/gpl-2.0.txt>.\n"
+	};
+#endif
+
+	static const char * const artists[] = {
+		"Icon derived from GTK+ Logo "
+		"https://wiki.gnome.org/Projects/GTK%2B/Logo",
+		NULL
 	};
 
 	static const char * const authors[] = {
-		"Tristan Heaven <tristanheaven@gmail.com>",
+		"Tristan Heaven <tristan@tristanheaven.net>",
 		NULL
 	};
 
 	gtk_show_about_dialog(
 			gui.window,
+			"artists", artists,
 			"authors", authors,
 			"comments", _("A GTK+ utility for computing message digests or checksums."),
+#if (GTK_MAJOR_VERSION > 2)
+			"license-type", GTK_LICENSE_GPL_2_0,
+#else
 			"license", license,
+#endif
+			"logo-icon-name", PACKAGE,
 			"program-name", PACKAGE_NAME,
 #if ENABLE_NLS
 			"translator-credits", _("translator-credits"),
 #endif
 			"version", VERSION,
-			"website", "http://gtkhash.sourceforge.net/",
+			"website", "https://github.com/tristanheaven/gtkhash",
 			NULL);
 }
 
 static void on_filechooserbutton_selection_changed(void)
 {
-	char *uri = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(gui.filechooserbutton));
+	bool enabled = hash_funcs_count_enabled();
+	char *uri = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(
+		gui.filechooserbutton));
 
-	gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash), uri ? true : false);
-
-	if (uri)
-		gtk_widget_grab_focus(GTK_WIDGET(gui.button_hash));
+	if (enabled && uri) {
+		g_free(uri);
+		gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash), true);
+	} else
+		gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash), false);
 
 	gui_clear_digests();
 }
@@ -261,34 +284,88 @@ static void on_toolbutton_add_clicked(void)
 	GtkFileChooser *chooser = GTK_FILE_CHOOSER(
 		gtk_file_chooser_dialog_new(_("Select Files"), gui.window,
 			GTK_FILE_CHOOSER_ACTION_OPEN,
-			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
-			GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+			_("_Cancel"), GTK_RESPONSE_CANCEL,
+			_("_Open"), GTK_RESPONSE_ACCEPT,
 			NULL));
 	gtk_file_chooser_set_select_multiple(chooser, true);
 	gtk_file_chooser_set_local_only(chooser, false);
 
 	if (gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_ACCEPT) {
 		GSList *uris = gtk_file_chooser_get_uris(chooser);
-		gui_add_uris(uris, GUI_VIEW_FILE_LIST);
-		g_slist_free_full(uris, g_free);
+		GSList *ud_list = uri_digest_list_from_uri_list(uris);
+
+		gui_add_ud_list(ud_list, GUI_VIEW_FILE_LIST);
+
+		uri_digest_list_free_full(ud_list);
+		g_slist_free(uris);
 	}
 
 	gtk_widget_destroy(GTK_WIDGET(chooser));
 }
 
-static void on_treeview_popup_menu(void)
+static void on_treeselection_changed(void)
 {
-	gtk_menu_popup(gui.menu_treeview, NULL, NULL, NULL, NULL, 0,
-		gtk_get_current_event_time());
+	const int rows = gtk_tree_selection_count_selected_rows(gui.treeselection);
+
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.toolbutton_remove), (rows > 0));
 }
 
-static bool on_treeview_button_press_event(G_GNUC_UNUSED GtkWidget *widget,
+static void show_menu_treeview(GdkEventButton *event)
+{
+	const int rows = gtk_tree_selection_count_selected_rows(gui.treeselection);
+
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_treeview_remove),
+		(rows > 0));
+
+	bool can_copy = false;
+
+	if (rows == 1) {
+		for (int i = 0; i < HASH_FUNCS_N; i++) {
+			if (!hash.funcs[i].enabled)
+				continue;
+			char *digest = list_get_selected_digest(i);
+			if (digest && *digest) {
+				can_copy = true;
+				gtk_widget_set_sensitive(GTK_WIDGET(
+					gui.hash_widgets[i].menuitem_treeview_copy), true);
+			} else {
+				gtk_widget_set_sensitive(GTK_WIDGET(
+					gui.hash_widgets[i].menuitem_treeview_copy), false);
+			}
+			g_free(digest);
+		}
+	}
+
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_treeview_copy),
+		can_copy);
+
+	if (event) {
+		gtk_menu_popup(gui.menu_treeview, NULL, NULL, NULL, NULL,
+			event->button, event->time);
+	} else {
+		gtk_menu_popup(gui.menu_treeview, NULL, NULL, NULL, NULL,
+			0, gtk_get_current_event_time());
+	}
+}
+
+static gboolean on_treeview_popup_menu(void)
+{
+	show_menu_treeview(NULL);
+	return true;
+}
+
+static gboolean on_treeview_button_press_event(G_GNUC_UNUSED GtkWidget *widget,
 	GdkEventButton *event)
 {
-	// Right click
+#if (GTK_MAJOR_VERSION > 2)
+	if (gdk_event_triggers_context_menu((GdkEvent *)event) &&
+		event->type == GDK_BUTTON_PRESS)
+	{
+#else
 	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
-		gtk_menu_popup(gui.menu_treeview, NULL, NULL, NULL, NULL, 3,
-			gdk_event_get_time((GdkEvent *)event));
+#endif
+		show_menu_treeview(event);
+		return true;
 	}
 
 	return false;
@@ -300,49 +377,21 @@ static void on_treeview_drag_data_received(G_GNUC_UNUSED GtkWidget *widget,
 	G_GNUC_UNUSED gpointer data)
 {
 	char **uris = gtk_selection_data_get_uris(selection);
-	GSList *list = NULL;
-
 	if (!uris) {
 		gtk_drag_finish(context, false, true, t);
 		return;
 	}
 
-	for (int i = 0; uris[i]; i++)
-		list = g_slist_prepend(list, uris[i]);
+	GSList *ud_list = uri_digest_list_from_uri_strv(uris);
 
-	list = g_slist_reverse(list);
+	gui_add_ud_list(ud_list, GUI_VIEW_FILE_LIST);
 
-	gui_add_uris(list, GUI_VIEW_FILE_LIST);
-
-	g_slist_free(list);
+	uri_digest_list_free(ud_list);
 	g_strfreev(uris);
 
 	gtk_drag_finish(context, true, true, t);
 }
 
-static void on_treeselection_changed(void)
-{
-	const int rows = gtk_tree_selection_count_selected_rows(gui.treeselection);
-
-	gtk_widget_set_sensitive(GTK_WIDGET(gui.toolbutton_remove), (rows > 0));
-
-	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_treeview_remove),
-		(rows > 0));
-	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_treeview_copy),
-		(rows == 1));
-
-	if (rows == 1) {
-		for (int i = 0; i < HASH_FUNCS_N; i++) {
-			if (!hash.funcs[i].enabled)
-				continue;
-			char *digest = list_get_selected_digest(i);
-			gtk_widget_set_sensitive(GTK_WIDGET(
-				gui.hash_widgets[i].menuitem_treeview_copy), (digest && *digest));
-			g_free(digest);
-		}
-	}
-}
-
 static void on_menuitem_treeview_copy_activate(struct hash_func_s *func)
 {
 	char *digest = list_get_selected_digest(func->id);
@@ -363,7 +412,7 @@ static void on_menuitem_treeview_show_toolbar_toggled(void)
 
 static void on_button_hash_clicked(void)
 {
-	if (gui_get_view() == GUI_VIEW_FILE) {
+	if (gui.view == GUI_VIEW_FILE) {
 		// XXX: Workaround for when user clicks Cancel in FileChooserDialog and
 		// XXX: uri is changed without emitting the "selection-changed" signal
 		on_filechooserbutton_selection_changed();
@@ -371,64 +420,57 @@ static void on_button_hash_clicked(void)
 			return;
 	}
 
-	switch (gui_get_view()) {
-		case GUI_VIEW_FILE: {
-			gui_clear_digests();
-			gui_set_state(GUI_STATE_BUSY);
-			char *uri = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(
-				gui.filechooserbutton));
-			hash_file_start(uri);
-			break;
-		}
-		case GUI_VIEW_TEXT:
-			hash_string();
-			break;
-		case GUI_VIEW_FILE_LIST:
-			gui_clear_digests();
-			gui_set_state(GUI_STATE_BUSY);
-			hash_file_list_start();
-			break;
-		default:
-			g_assert_not_reached();
-	}
+	gui_start_hash();
 }
 
-static void on_button_stop_clicked(void)
+static void on_togglebutton_hmac_file_toggled(void)
 {
-	gtk_widget_set_sensitive(GTK_WIDGET(gui.button_stop), false);
-	hash_file_stop();
+	g_assert(gui.view == GUI_VIEW_FILE);
+
+	bool active = gtk_toggle_button_get_active(gui.togglebutton_hmac_file);
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_file), active);
+	gui_clear_digests();
+
+	gui_update_hash_func_labels(active);
 }
 
-static void on_entry_hmac_changed(void)
+static void on_togglebutton_hmac_text_toggled(void)
 {
-	switch (gui_get_view()) {
-		case GUI_VIEW_FILE:
-			gui_clear_digests();
-			break;
-		case GUI_VIEW_TEXT:
-			on_button_hash_clicked();
-			break;
-		default:
-			g_assert_not_reached();
-	}
+	g_assert(gui.view == GUI_VIEW_TEXT);
+
+	bool active = gtk_toggle_button_get_active(gui.togglebutton_hmac_text);
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_text), active);
+
+	hash_string();
+
+	gui_update_hash_func_labels(active);
 }
 
-static void on_togglebutton_hmac_toggled(void)
+static void on_menuitem_show_hmac_key_toggled(GtkCheckMenuItem *item,
+	GtkEntry *entry)
 {
-	switch (gui_get_view()) {
-		case GUI_VIEW_FILE:
-			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_file),
-				gtk_toggle_button_get_active(gui.togglebutton_hmac_file));
-			break;
-		case GUI_VIEW_TEXT:
-			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_text),
-				gtk_toggle_button_get_active(gui.togglebutton_hmac_text));
-			break;
-		default:
-			g_assert_not_reached();
-	}
+	const bool active = gtk_check_menu_item_get_active(item);
 
-	on_entry_hmac_changed();
+	gtk_entry_set_visibility(entry, active);
+}
+
+static void on_entry_hmac_populate_popup(GtkEntry *entry, GtkMenu *menu)
+{
+	GtkWidget *item;
+
+	// Add separator
+	item = gtk_separator_menu_item_new();
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
+	// Add checkbutton
+	item = gtk_check_menu_item_new_with_mnemonic(_("_Show HMAC Key"));
+	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),
+		gtk_entry_get_visibility(entry));
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+	g_signal_connect(item, "toggled",
+		G_CALLBACK(on_menuitem_show_hmac_key_toggled), entry);
 }
 
 static bool on_dialog_delete_event(void)
@@ -439,16 +481,29 @@ static bool on_dialog_delete_event(void)
 
 static void on_dialog_combobox_changed(void)
 {
-	gui_clear_all_digests();
+	for (int i = 0; i < HASH_FUNCS_N; i++) {
+		if (!hash.funcs[i].supported)
+			continue;
 
-	if (gui_get_view() == GUI_VIEW_TEXT)
-		on_button_hash_clicked();
+		gtk_entry_set_text(gui.hash_widgets[i].entry_file, "");
+		gtk_entry_set_text(gui.hash_widgets[i].entry_text, "");
+	}
+
+	list_clear_digests();
+
+	if (gui.view == GUI_VIEW_TEXT)
+		hash_string();
+	else
+		gui_check_digests();
 }
 
 void callbacks_init(void)
 {
-#define CON(OBJ, SIG, CB) g_signal_connect(G_OBJECT(OBJ), SIG, CB, NULL)
-	CON(gui.window,                         "delete-event",        G_CALLBACK(on_window_delete_event));
+#define CON(OBJ, SIG, CB) \
+	g_signal_connect(G_OBJECT(OBJ), SIG, G_CALLBACK(CB), NULL)
+
+	CON(gui.window,                         "delete-event",        on_window_delete_event);
+	CON(gui.menuitem_open,                  "activate",            on_menuitem_open_activate);
 	CON(gui.menuitem_save_as,               "activate",            on_menuitem_save_as_activate);
 	CON(gui.menuitem_quit,                  "activate",            on_menuitem_quit_activate);
 	CON(gui.menuitem_edit,                  "activate",            on_menuitem_edit_activate);
@@ -458,38 +513,43 @@ void callbacks_init(void)
 	CON(gui.menuitem_delete,                "activate",            on_menuitem_delete_activate);
 	CON(gui.menuitem_select_all,            "activate",            on_menuitem_select_all_activate);
 	CON(gui.menuitem_prefs,                 "activate",            on_menuitem_prefs_activate);
-	CON(gui.radiomenuitem_file,             "toggled",             gui_update);
-	CON(gui.radiomenuitem_text,             "toggled",             gui_update);
-	CON(gui.radiomenuitem_file_list,        "toggled",             gui_update);
+	CON(gui.radiomenuitem_file,             "toggled",             on_radiomenuitem_toggled);
+	CON(gui.radiomenuitem_text,             "toggled",             on_radiomenuitem_toggled);
+	CON(gui.radiomenuitem_file_list,        "toggled",             on_radiomenuitem_toggled);
 	CON(gui.menuitem_about,                 "activate",            on_menuitem_about_activate);
 //	file-set isn't emitted when file is deleted
 //	CON(gui.filechooserbutton,              "file-set",            on_filechooserbutton_file_set);
 	CON(gui.filechooserbutton,              "selection-changed",   on_filechooserbutton_selection_changed);
-	CON(gui.entry_text,                     "changed",             on_button_hash_clicked);
-	CON(gui.togglebutton_hmac_file,         "toggled",             on_togglebutton_hmac_toggled);
-	CON(gui.togglebutton_hmac_text,         "toggled",             on_togglebutton_hmac_toggled);
-	CON(gui.entry_hmac_file,                "changed",             on_entry_hmac_changed);
-	CON(gui.entry_hmac_text,                "changed",             on_entry_hmac_changed);
+	CON(gui.entry_text,                     "changed",             hash_string);
+	CON(gui.togglebutton_hmac_file,         "toggled",             on_togglebutton_hmac_file_toggled);
+	CON(gui.togglebutton_hmac_text,         "toggled",             on_togglebutton_hmac_text_toggled);
+	CON(gui.entry_hmac_file,                "populate-popup",      on_entry_hmac_populate_popup);
+	CON(gui.entry_hmac_text,                "populate-popup",      on_entry_hmac_populate_popup);
+	CON(gui.entry_hmac_file,                "changed",             gui_clear_digests);
+	CON(gui.entry_hmac_text,                "changed",             hash_string);
 	CON(gui.entry_check_file,               "changed",             gui_check_digests);
 	CON(gui.entry_check_text,               "changed",             gui_check_digests);
 	CON(gui.toolbutton_add,                 "clicked",             on_toolbutton_add_clicked);
 	CON(gui.toolbutton_remove,              "clicked",             list_remove_selection);
 	CON(gui.toolbutton_clear,               "clicked",             list_clear);
-	CON(gui.treeview,                       "popup-menu",          on_treeview_popup_menu);
-	CON(gui.treeview,                       "button-press-event",  G_CALLBACK(on_treeview_button_press_event));
-	CON(gui.treeview,                       "drag-data-received",  G_CALLBACK(on_treeview_drag_data_received));
 	CON(gui.treeselection,                  "changed",             on_treeselection_changed);
+	CON(gui.treeview,                       "popup-menu",          on_treeview_popup_menu);
+	CON(gui.treeview,                       "button-press-event",  on_treeview_button_press_event);
+	CON(gui.treeview,                       "drag-data-received",  on_treeview_drag_data_received);
 	CON(gui.menuitem_treeview_add,          "activate",            on_toolbutton_add_clicked);
 	CON(gui.menuitem_treeview_remove,       "activate",            list_remove_selection);
 	CON(gui.menuitem_treeview_clear,        "activate",            list_clear);
 	CON(gui.menuitem_treeview_show_toolbar, "toggled",             on_menuitem_treeview_show_toolbar_toggled);
 	CON(gui.button_hash,                    "clicked",             on_button_hash_clicked);
-	CON(gui.button_stop,                    "clicked",             on_button_stop_clicked);
+	CON(gui.button_stop,                    "clicked",             gui_stop_hash);
 	CON(gui.dialog,                         "delete-event",        G_CALLBACK(on_dialog_delete_event));
 	CON(gui.dialog_button_close,            "clicked",             G_CALLBACK(on_dialog_delete_event));
 	CON(gui.dialog_combobox,                "changed",             on_dialog_combobox_changed);
 
 	for (int i = 0; i < HASH_FUNCS_N; i++) {
+		if (!hash.funcs[i].supported)
+			continue;
+
 		CON(gui.hash_widgets[i].button, "toggled", gui_update);
 		g_signal_connect_swapped(gui.hash_widgets[i].menuitem_treeview_copy,
 			"activate", G_CALLBACK(on_menuitem_treeview_copy_activate),
