@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2007-2011 Tristan Heaven <tristanheaven@gmail.com>
+ *   Copyright (C) 2007-2016 Tristan Heaven <tristan@tristanheaven.net>
  *
  *   This file is part of GtkHash.
  *
@@ -14,7 +14,7 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with GtkHash. If not, see <http://www.gnu.org/licenses/>.
+ *   along with GtkHash. If not, see <https://gnu.org/licenses/gpl-2.0.txt>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -34,9 +34,19 @@
 #include "hash.h"
 #include "list.h"
 #include "prefs.h"
+#include "uri-digest.h"
 #include "hash/digest-format.h"
+#include "hash/hash-func.h"
 
-struct gui_s gui;
+#if (GTK_MAJOR_VERSION > 2)
+#define GUI_XML_RESOURCE "/org/gtkhash/gtkhash-gtk3.xml"
+#else
+#define GUI_XML_RESOURCE "/org/gtkhash/gtkhash-gtk2.xml"
+#endif
+
+struct gui_s gui = {
+	.view = GUI_VIEW_INVALID,
+};
 
 static struct {
 	enum gui_state_e state;
@@ -55,15 +65,15 @@ static GObject *gui_get_object(GtkBuilder *builder, const char *name)
 	return obj;
 }
 
-static void gui_get_objects(GtkBuilder *builder)
+static void gui_init_objects(GtkBuilder *builder)
 {
 	// Window
 	gui.window = GTK_WINDOW(gui_get_object(builder,
 		"window"));
 
 	// Menus
-	gui.menuitem_file = GTK_MENU_ITEM(gui_get_object(builder,
-		"menuitem_file"));
+	gui.menuitem_open = GTK_MENU_ITEM(gui_get_object(builder,
+		"menuitem_open"));
 	gui.menuitem_save_as = GTK_MENU_ITEM(gui_get_object(builder,
 		"menuitem_save_as"));
 	gui.menuitem_quit = GTK_MENU_ITEM(gui_get_object(builder,
@@ -148,6 +158,7 @@ static void gui_get_objects(GtkBuilder *builder)
 		"treeselection"));
 	gui.menu_treeview = GTK_MENU(gui_get_object(builder,
 		"menu_treeview"));
+	g_object_ref(gui.menu_treeview);
 	gui.menuitem_treeview_add = GTK_MENU_ITEM(gui_get_object(builder,
 		"menuitem_treeview_add"));
 	gui.menuitem_treeview_remove = GTK_MENU_ITEM(gui_get_object(builder,
@@ -176,8 +187,13 @@ static void gui_get_objects(GtkBuilder *builder)
 	// Dialog
 	gui.dialog = GTK_DIALOG(gui_get_object(builder,
 		"dialog"));
+#if (GTK_MAJOR_VERSION > 2)
+	gui.dialog_grid = GTK_GRID(gui_get_object(builder,
+		"dialog_grid"));
+#else
 	gui.dialog_table = GTK_TABLE(gui_get_object(builder,
 		"dialog_table"));
+#endif
 	gui.dialog_combobox = GTK_COMBO_BOX(gui_get_object(builder,
 		"dialog_combobox"));
 	gui.dialog_button_close = GTK_BUTTON(gui_get_object(builder,
@@ -186,18 +202,22 @@ static void gui_get_objects(GtkBuilder *builder)
 
 static void gui_init_hash_funcs(void)
 {
-	int supported = 0;
+	for (int i = 0, supported = 0; i < HASH_FUNCS_N; i++) {
+		if (!hash.funcs[i].supported)
+			continue;
 
-	for (int i = 0; i < HASH_FUNCS_N; i++) {
 		// File/Text view func labels
-		char *label = g_strdup_printf("%s:", hash.funcs[i].name);
-		gui.hash_widgets[i].label = GTK_LABEL(gtk_label_new(label));
-		g_free(label);
+		gui.hash_widgets[i].label = GTK_LABEL(gtk_label_new(NULL));
 		gtk_container_add(GTK_CONTAINER(gui.vbox_outputlabels),
 			GTK_WIDGET(gui.hash_widgets[i].label));
+#if (GTK_MAJOR_VERSION > 2)
+		gtk_widget_set_halign(GTK_WIDGET(gui.hash_widgets[i].label),
+			GTK_ALIGN_START);
+#else
 		gtk_misc_set_alignment(GTK_MISC(gui.hash_widgets[i].label),
 			// Left align
 			0.0, 0.5);
+#endif
 
 		// File view digests
 		gui.hash_widgets[i].entry_file = GTK_ENTRY(gtk_entry_new());
@@ -226,122 +246,114 @@ static void gui_init_hash_funcs(void)
 		// Dialog checkbuttons
 		gui.hash_widgets[i].button = GTK_TOGGLE_BUTTON(
 			gtk_check_button_new_with_label(hash.funcs[i].name));
-		if (hash.funcs[i].supported) {
-			gtk_table_attach_defaults(gui.dialog_table,
-				GTK_WIDGET(gui.hash_widgets[i].button),
-				// Sort checkbuttons into 2 columns
-				supported % 2 ? 1 : 0,
-				supported % 2 ? 2 : 1,
-				supported / 2,
-				supported / 2 + 1);
-			gtk_widget_show(GTK_WIDGET(gui.hash_widgets[i].button));
-
-			supported++;
-		}
-	}
-}
-
-static char *gui_get_xml(const char *filename)
-{
-	GMappedFile *map = g_mapped_file_new(filename, false, NULL);
-
-	if (!map)
-		return NULL;
-
-	gsize map_len = g_mapped_file_get_length(map);
-	if (map_len == 0) {
-		g_mapped_file_unref(map);
-		return NULL;
-	}
-
-	const char *map_data = g_mapped_file_get_contents(map);
-	g_assert(map_data);
-
-	GInputStream *input_mem = g_memory_input_stream_new_from_data(map_data,
-		map_len, NULL);
-	GConverter *converter = G_CONVERTER(g_zlib_decompressor_new(
-		G_ZLIB_COMPRESSOR_FORMAT_GZIP));
-
-	GInputStream *input_conv = g_converter_input_stream_new(input_mem,
-		converter);
-
-	g_object_unref(input_mem);
-	g_object_unref(converter);
-
-	GString *string = g_string_new(NULL);
-
-	for (char buf[1024];;) {
-		gssize len = g_input_stream_read(input_conv, buf, 1024, NULL, NULL);
-		if (len <= 0)
-			break;
-
-		g_string_append_len(string, buf, len);
-	}
-
-	g_object_unref(input_conv);
-	g_mapped_file_unref(map);
-
-	return g_string_free(string, false);
-}
-
-void gui_init(const char *datadir)
-{
-	gtk_init(NULL, NULL);
-
-	char *filename = g_build_filename(datadir, PACKAGE ".xml.gz", NULL);
-	char *xml = gui_get_xml(filename);
-
-	if (!xml || !*xml) {
-		char *message = g_strdup_printf(_("Failed to read \"%s\""), filename);
-		gui_error(message);
-		g_free(message);
-		g_free(xml);
-		g_free(filename);
-		exit(EXIT_FAILURE);
-	}
-
-	GtkBuilder *builder = gtk_builder_new();
-	gtk_builder_set_translation_domain(builder, GETTEXT_PACKAGE);
-
-	GError *error = NULL;
-	gtk_builder_add_from_string(builder, xml, -1, &error);
-
-	g_free(xml);
-
-	if (error) {
-		char *message = g_strdup_printf(_("Failed to read \"%s\":\n%s"),
-			filename, error->message);
-		gui_error(message);
-		g_free(message);
-		g_free(filename);
-		g_error_free(error);
-		g_object_unref(builder);
-		exit(EXIT_FAILURE);
-	}
-
-	g_free(filename);
-
-	gui_get_objects(builder);
-	g_object_ref(gui.menu_treeview);
-	g_object_unref(builder);
 
 #if (GTK_MAJOR_VERSION > 2)
-	// Avoid GTK+ 2 "Unknown property" warning
-	gtk_widget_set_valign(GTK_WIDGET(gui.vbox_single), GTK_ALIGN_START);
+		gtk_grid_attach(gui.dialog_grid,
+			GTK_WIDGET(gui.hash_widgets[i].button),
+			supported % 2 ? 1 : 0, // column
+			supported / 2,         // row
+			1,                     // width
+			1);                    // height
+#else
+		gtk_table_attach_defaults(gui.dialog_table,
+			GTK_WIDGET(gui.hash_widgets[i].button),
+			// Sort checkbuttons into 2 columns
+			supported % 2 ? 1 : 0,
+			supported % 2 ? 2 : 1,
+			supported / 2,
+			supported / 2 + 1);
 #endif
+		// Could be enabled already by cmdline arg
+		if (hash.funcs[i].enabled)
+			gtk_toggle_button_set_active(gui.hash_widgets[i].button, true);
+
+		gtk_widget_show(GTK_WIDGET(gui.hash_widgets[i].button));
+
+		supported++;
+	}
+}
+
+#if (GTK_MAJOR_VERSION < 3)
+static void gui_init_fonts(void)
+{
+	PangoFontDescription *desc = pango_font_description_from_string(
+		"monospace");
+
+	gtk_widget_modify_font(GTK_WIDGET(gui.entry_check_file), desc);
+	gtk_widget_modify_font(GTK_WIDGET(gui.entry_check_text), desc);
+	gtk_widget_modify_font(GTK_WIDGET(gui.treeview), desc);
+
+	for (int i = 0; i < HASH_FUNCS_N; i++) {
+		if (!hash.funcs[i].supported)
+			continue;
+
+		gtk_widget_modify_font(GTK_WIDGET(gui.hash_widgets[i].entry_file), desc);
+		gtk_widget_modify_font(GTK_WIDGET(gui.hash_widgets[i].entry_text), desc);
+	}
+
+	pango_font_description_free(desc);
+}
+#endif
+
+static GtkBuilder *gui_init_builder(void)
+{
+#if (GTK_MAJOR_VERSION > 2)
+	return gtk_builder_new_from_resource(GUI_XML_RESOURCE);
+#else
+	GError *error = NULL;
+	GBytes *bytes = g_resources_lookup_data(GUI_XML_RESOURCE,
+		G_RESOURCE_LOOKUP_FLAGS_NONE, &error);
+
+	if (G_UNLIKELY(error)) {
+		gui_error(error->message);
+		g_error_free(error);
+
+		exit(EXIT_FAILURE);
+	}
+
+	gsize xml_len = 0;
+	char *xml = g_bytes_unref_to_data(bytes, &xml_len);
+	GtkBuilder *builder = gtk_builder_new();
+
+	gtk_builder_add_from_string(builder, xml, xml_len, &error);
+	g_free(xml);
+
+	if (G_UNLIKELY(error)) {
+		gui_error(error->message);
+		g_error_free(error);
+		g_object_unref(builder);
+
+		exit(EXIT_FAILURE);
+	}
+
+	return builder;
+#endif
+}
+
+void gui_init(void)
+{
+	GtkBuilder *builder = gui_init_builder();
+	gui_init_objects(builder);
+	g_object_unref(builder);
 
 	gui_init_hash_funcs();
 
-	gui_set_state(GUI_STATE_IDLE);
+#if (GTK_MAJOR_VERSION < 3)
+	gui_init_fonts();
 
-	callbacks_init();
+	g_object_set(gtk_settings_get_default(), "gtk-button-images", true, NULL);
+	g_object_set(gtk_settings_get_default(), "gtk-menu-images", true, NULL);
+#endif
+
+	gui_set_state(GUI_STATE_IDLE);
 }
 
-static bool gui_can_add_uri(char *uri, char **error_str)
+static char *gui_try_uri(const char *uri)
 {
 	g_assert(uri);
 
-	bool can_add = false;
+	char *error_str = NULL;
+
 	GFile *file = g_file_new_for_uri(uri);
 	GError *error = NULL;
 	GFileInfo *info = g_file_query_info(file,
@@ -354,66 +366,64 @@ static bool gui_can_add_uri(char *uri, char **error_str)
 		GFileType type = g_file_info_get_file_type(info);
 		g_object_unref(info);
 		if (!can_read)
-			*error_str = g_strdup(g_strerror(EACCES));
-		else if (type == G_FILE_TYPE_DIRECTORY)
-			*error_str = g_strdup(g_strerror(EISDIR));
+			error_str = g_strdup(g_strerror(EACCES));
+		else if (type == G_FILE_TYPE_DIRECTORY) // TODO
+			error_str = g_strdup(g_strerror(EISDIR));
 		else if (type != G_FILE_TYPE_REGULAR)
-			*error_str = g_strdup(_("Not a regular file"));
-		else
-			can_add = true;
+			error_str = g_strdup(_("Not a regular file"));
 	} else {
-		*error_str = g_strdup(error->message);
+		error_str = g_strdup(error->message);
 		g_error_free(error);
 	}
 
 	g_object_unref(file);
 
-	return can_add;
+	return error_str;
 }
 
-unsigned int gui_add_uris(GSList *uris, enum gui_view_e view)
+unsigned int gui_add_ud_list(GSList *ud_list, const enum gui_view_e view)
 {
-	g_assert(uris);
+	g_assert(ud_list);
 
 	GSList *readable = NULL;
 	unsigned int readable_len = 0;
 	{
-		GSList *tmp = uris;
+		GSList *tmp = ud_list;
 		do {
-			char *error = g_strdup(_("Unknown error"));
-			if (!gui_can_add_uri(tmp->data, &error)) {
-				char *message = g_strdup_printf(_("Failed to add \"%s\":\n%s"),
-					(char *)tmp->data, error);
-				gui_error(message);
+			char *error = NULL;
+			struct uri_digest_s *ud = tmp->data;
+			if ((error = gui_try_uri(ud->uri))) {
+				g_message(_("Failed to add file \"%s\": %s"),
+					ud->uri, error);
 				g_free(error);
-				g_free(message);
 				continue;
 			}
-			if (!g_slist_find_custom(readable, tmp->data,
-				(GCompareFunc)g_strcmp0))
-			{
-				readable = g_slist_prepend(readable, tmp->data);
-				readable_len++;
-			}
+			readable = g_slist_prepend(readable, tmp->data);
+			readable_len++;
 		} while ((tmp = tmp->next));
 	}
 	readable = g_slist_reverse(readable);
 
-	if (view == GUI_VIEW_INVALID) {
+	if (!GUI_VIEW_IS_VALID(view)) {
 		if (readable_len == 1)
-			view = GUI_VIEW_FILE;
+			gui_set_view(GUI_VIEW_FILE);
 		else if (readable_len > 1)
-			view = GUI_VIEW_FILE_LIST;
+			gui_set_view(GUI_VIEW_FILE_LIST);
 	}
 
-	if (readable_len && (view == GUI_VIEW_FILE)) {
+	if (readable_len && (gui.view == GUI_VIEW_FILE)) {
+		struct uri_digest_s *ud = readable->data;
 		gtk_file_chooser_set_uri(GTK_FILE_CHOOSER(gui.filechooserbutton),
-			readable->data);
-	} else if (readable_len && (view == GUI_VIEW_FILE_LIST)) {
+			ud->uri);
+		if (ud->digest)
+			gtk_entry_set_text(gui.entry_check_file, ud->digest);
+	} else if (readable_len && (gui.view == GUI_VIEW_FILE_LIST)) {
 		GSList *tmp = readable;
 		do {
-			list_append_row(tmp->data);
+			struct uri_digest_s *ud = tmp->data;
+			list_append_row(ud->uri, ud->digest);
 		} while ((tmp = tmp->next));
+		list_update();
 	}
 
 	g_slist_free(readable);
@@ -421,6 +431,26 @@ unsigned int gui_add_uris(GSList *uris, enum gui_view_e view)
 	return readable_len;
 }
 
+void gui_add_check(const char *check)
+{
+	g_assert(check && *check);
+
+	gtk_entry_set_text(gui.entry_check_file, check);
+	gtk_entry_set_text(gui.entry_check_text, check);
+}
+
+void gui_add_text(const char *text)
+{
+	g_assert(text);
+
+	gtk_entry_set_text(gui.entry_text, text);
+
+	gui_set_view(GUI_VIEW_TEXT);
+	gui_update();
+
+	gtk_editable_set_position(GTK_EDITABLE(gui.entry_text), -1);
+}
+
 void gui_error(const char *message)
 {
 	GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
@@ -432,8 +462,16 @@ void gui_error(const char *message)
 
 void gui_run(void)
 {
-	// Show window here so it isn't resized just after it's opened
-	gtk_widget_show(GTK_WIDGET(gui.window));
+	// Set default view
+	if (!GUI_VIEW_IS_VALID(gui.view)) {
+		gui_set_view(GUI_VIEW_FILE);
+		gui_update();
+	}
+
+	gtk_widget_show_now(GTK_WIDGET(gui.window));
+
+	// Connect signals to start handling events
+	callbacks_init();
 
 	gtk_main();
 }
@@ -442,7 +480,7 @@ void gui_deinit(void)
 {
 	hash_file_stop();
 
-	while (gui_get_state() != GUI_STATE_IDLE)
+	while (gui_priv.state != GUI_STATE_IDLE)
 		gtk_main_iteration();
 
 	gtk_widget_destroy(GTK_WIDGET(gui.window));
@@ -454,6 +492,13 @@ void gui_deinit(void)
 
 void gui_set_view(const enum gui_view_e view)
 {
+	g_assert(GUI_VIEW_IS_VALID(view));
+
+	if (view == gui.view)
+		return;
+
+	gui.view = view;
+
 	switch (view) {
 		case GUI_VIEW_FILE:
 			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(
@@ -472,30 +517,6 @@ void gui_set_view(const enum gui_view_e view)
 	}
 }
 
-enum gui_view_e gui_get_view(void)
-{
-	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
-		gui.radiomenuitem_file)))
-	{
-		return GUI_VIEW_FILE;
-	}
-
-	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
-		gui.radiomenuitem_text)))
-	{
-		return GUI_VIEW_TEXT;
-	}
-
-	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
-		gui.radiomenuitem_file_list)))
-	{
-		return GUI_VIEW_FILE_LIST;
-	}
-
-	g_assert_not_reached();
-	return GUI_VIEW_INVALID;
-}
-
 void gui_set_digest_format(const enum digest_format_e format)
 {
 	g_assert(DIGEST_FORMAT_IS_VALID(format));
@@ -518,21 +539,26 @@ const uint8_t *gui_get_hmac_key(size_t *key_size)
 	const uint8_t *hmac_key = NULL;
 	*key_size = 0;
 
-	switch (gui_get_view()) {
+	switch (gui.view) {
 		case GUI_VIEW_FILE:
 			if (gtk_toggle_button_get_active(gui.togglebutton_hmac_file)) {
+				GtkEntryBuffer *buffer = gtk_entry_get_buffer(
+					gui.entry_hmac_file);
+
 				hmac_key = (uint8_t *)gtk_entry_get_text(gui.entry_hmac_file);
-				*key_size = gtk_entry_get_text_length(gui.entry_hmac_file);
+				*key_size = gtk_entry_buffer_get_bytes(buffer);
 			}
 			break;
 		case GUI_VIEW_TEXT:
 			if (gtk_toggle_button_get_active(gui.togglebutton_hmac_text)) {
+				GtkEntryBuffer *buffer = gtk_entry_get_buffer(
+					gui.entry_hmac_text);
+
 				hmac_key = (uint8_t *)gtk_entry_get_text(gui.entry_hmac_text);
-				*key_size = gtk_entry_get_text_length(gui.entry_hmac_text);
+				*key_size = gtk_entry_buffer_get_bytes(buffer);
 			}
 			break;
-		case GUI_VIEW_FILE_LIST:
-			// TODO
+		case GUI_VIEW_FILE_LIST: // TODO
 			break;
 		default:
 			g_assert_not_reached();
@@ -541,18 +567,88 @@ const uint8_t *gui_get_hmac_key(size_t *key_size)
 	return hmac_key;
 }
 
-void gui_update(void)
+static void gui_menuitem_save_as_set_sensitive(void)
 {
-	bool has_enabled = false;
+	if (gui_priv.state == GUI_STATE_BUSY) {
+		gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_save_as), false);
+		return;
+	}
 
+	bool sensitive = false;
+
+	switch (gui.view) {
+		case GUI_VIEW_FILE:
+			for (int i = 0; i < HASH_FUNCS_N; i++) {
+				if (hash.funcs[i].enabled &&
+					*gtk_entry_get_text(gui.hash_widgets[i].entry_file))
+				{
+					sensitive = true;
+					break;
+				}
+			}
+			break;
+		case GUI_VIEW_TEXT:
+			for (int i = 0; i < HASH_FUNCS_N; i++) {
+				if (hash.funcs[i].enabled) {
+					sensitive = true;
+					break;
+				}
+			}
+			break;
+		case GUI_VIEW_FILE_LIST:
+			for (int i = 0; i < HASH_FUNCS_N; i++) {
+				if (hash.funcs[i].enabled) {
+					char *digest = list_get_digest(0, i);
+					if (digest != NULL && *digest) {
+						g_free(digest);
+						sensitive = true;
+						break;
+					}
+				}
+			}
+			break;
+		default:
+			sensitive = false;
+	}
+
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_save_as), sensitive);
+}
+
+void gui_enable_hash_func(const enum hash_func_e id)
+{
+	g_assert(HASH_FUNC_IS_VALID(id));
+
+	if (hash.funcs[id].supported)
+		gtk_toggle_button_set_active(gui.hash_widgets[id].button, true);
+}
+
+void gui_update_hash_func_labels(const bool hmac_enabled)
+{
 	for (int i = 0; i < HASH_FUNCS_N; i++) {
-		if (hash.funcs[i].supported) {
-			hash.funcs[i].enabled = gtk_toggle_button_get_active(
-				gui.hash_widgets[i].button);
-			if (hash.funcs[i].enabled)
-				has_enabled = true;
-		} else
-			hash.funcs[i].enabled = false;
+		if (!hash.funcs[i].enabled)
+			continue;
+
+		char *str = NULL;
+
+		// FIXME: different labels for RTL?
+		if (hmac_enabled && hash.funcs[i].hmac_supported)
+			str = g_strdup_printf("HMAC-%s:", hash.funcs[i].name);
+		else
+			str = g_strdup_printf("%s:", hash.funcs[i].name);
+
+		gtk_label_set_text(gui.hash_widgets[i].label, str);
+		g_free(str);
+	}
+}
+
+static void gui_update_hash_funcs(void)
+{
+	for (int i = 0; i < HASH_FUNCS_N; i++) {
+		if (!hash.funcs[i].supported)
+			continue;
+
+		hash.funcs[i].enabled = gtk_toggle_button_get_active(
+			gui.hash_widgets[i].button);
 
 		gtk_widget_set_visible(GTK_WIDGET(gui.hash_widgets[i].label),
 			hash.funcs[i].enabled);
@@ -563,42 +659,74 @@ void gui_update(void)
 		gtk_widget_set_visible(GTK_WIDGET(gui.hash_widgets[i].menuitem_treeview_copy),
 			hash.funcs[i].enabled);
 	}
+}
 
-	list_update();
+static void gui_update_hmac(void)
+{
+	bool active = false;
 
-	enum gui_view_e view = gui_get_view();
+	switch (gui.view) {
+		case GUI_VIEW_FILE:
+			gtk_widget_hide(GTK_WIDGET(gui.entry_hmac_text));
+			gtk_widget_hide(GTK_WIDGET(gui.togglebutton_hmac_text));
+			gtk_widget_show(GTK_WIDGET(gui.entry_hmac_file));
+			gtk_widget_show(GTK_WIDGET(gui.togglebutton_hmac_file));
+
+			active = gtk_toggle_button_get_active(gui.togglebutton_hmac_file);
+			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_file), active);
+			break;
+		case GUI_VIEW_TEXT:
+			gtk_widget_hide(GTK_WIDGET(gui.entry_hmac_file));
+			gtk_widget_hide(GTK_WIDGET(gui.togglebutton_hmac_file));
+			gtk_widget_show(GTK_WIDGET(gui.entry_hmac_text));
+			gtk_widget_show(GTK_WIDGET(gui.togglebutton_hmac_text));
+
+			active = gtk_toggle_button_get_active(gui.togglebutton_hmac_text);
+			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_text), active);
+			break;
+		default:
+			g_assert_not_reached();
+	}
+
+	gui_update_hash_func_labels(active);
+}
+
+void gui_update(void)
+{
+	// Must call gui_set_view() before this
+	g_assert(GUI_VIEW_IS_VALID(gui.view));
+	g_assert(gui_priv.state == GUI_STATE_IDLE);
+
+	gui_update_hash_funcs();
+	const unsigned int funcs_enabled = hash_funcs_count_enabled();
+
+	if ((gui.view == GUI_VIEW_FILE) || (gui.view == GUI_VIEW_TEXT)) {
+		gui_update_hmac();
 
-	if ((view == GUI_VIEW_FILE) || (view == GUI_VIEW_TEXT)) {
 		gtk_widget_hide(GTK_WIDGET(gui.toolbar));
 		gtk_widget_hide(GTK_WIDGET(gui.vbox_list));
 		gtk_widget_show(GTK_WIDGET(gui.vbox_single));
 	}
 
-	switch (view) {
+	switch (gui.view) {
 		case GUI_VIEW_FILE: {
 			gtk_widget_hide(GTK_WIDGET(gui.label_text));
 			gtk_widget_hide(GTK_WIDGET(gui.entry_text));
 			gtk_widget_hide(GTK_WIDGET(gui.entry_check_text));
-			gtk_widget_hide(GTK_WIDGET(gui.entry_hmac_text));
-			gtk_widget_hide(GTK_WIDGET(gui.togglebutton_hmac_text));
+
 			gtk_widget_hide(GTK_WIDGET(gui.vbox_digests_text));
 			gtk_widget_show(GTK_WIDGET(gui.label_file));
 			gtk_widget_show(GTK_WIDGET(gui.filechooserbutton));
 			gtk_widget_show(GTK_WIDGET(gui.entry_check_file));
-			gtk_widget_show(GTK_WIDGET(gui.entry_hmac_file));
-			gtk_widget_show(GTK_WIDGET(gui.togglebutton_hmac_file));
 			gtk_widget_show(GTK_WIDGET(gui.vbox_digests_file));
 			gtk_widget_show(GTK_WIDGET(gui.hseparator_buttons));
 			gtk_widget_show(GTK_WIDGET(gui.button_hash));
 
-			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_file),
-				gtk_toggle_button_get_active(gui.togglebutton_hmac_file));
-
 			char *uri = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(
 				gui.filechooserbutton));
 			if (uri) {
 				gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash),
-					has_enabled);
+					funcs_enabled);
 				g_free(uri);
 			} else
 				gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash), false);
@@ -608,26 +736,22 @@ void gui_update(void)
 			gtk_widget_hide(GTK_WIDGET(gui.label_file));
 			gtk_widget_hide(GTK_WIDGET(gui.filechooserbutton));
 			gtk_widget_hide(GTK_WIDGET(gui.entry_check_file));
-			gtk_widget_hide(GTK_WIDGET(gui.entry_hmac_file));
-			gtk_widget_hide(GTK_WIDGET(gui.togglebutton_hmac_file));
 			gtk_widget_hide(GTK_WIDGET(gui.vbox_digests_file));
 			gtk_widget_hide(GTK_WIDGET(gui.hseparator_buttons));
 			gtk_widget_hide(GTK_WIDGET(gui.button_hash));
 			gtk_widget_show(GTK_WIDGET(gui.label_text));
 			gtk_widget_show(GTK_WIDGET(gui.entry_text));
 			gtk_widget_show(GTK_WIDGET(gui.entry_check_text));
-			gtk_widget_show(GTK_WIDGET(gui.entry_hmac_text));
-			gtk_widget_show(GTK_WIDGET(gui.togglebutton_hmac_text));
 			gtk_widget_show(GTK_WIDGET(gui.vbox_digests_text));
 
-			gtk_widget_set_sensitive(GTK_WIDGET(gui.entry_hmac_file),
-				gtk_toggle_button_get_active(gui.togglebutton_hmac_file));
-
 			gtk_widget_grab_focus(GTK_WIDGET(gui.entry_text));
 
-			gtk_button_clicked(gui.button_hash);
+			if (funcs_enabled)
+				hash_string();
 			break;
 		case GUI_VIEW_FILE_LIST:
+			list_update();
+
 			gtk_widget_hide(GTK_WIDGET(gui.vbox_single));
 			gtk_widget_hide(GTK_WIDGET(gui.hseparator_buttons));
 			gtk_widget_show(GTK_WIDGET(gui.vbox_list));
@@ -636,27 +760,32 @@ void gui_update(void)
 			gtk_widget_set_visible(GTK_WIDGET(gui.toolbar),
 				gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(
 					gui.menuitem_treeview_show_toolbar)));
-			gtk_widget_set_sensitive(GTK_WIDGET(gui.button_hash),
-				has_enabled && (list_count_rows() > 0));
 			break;
 		default:
 			g_assert_not_reached();
 	}
 
 	gui_check_digests();
+	gui_menuitem_save_as_set_sensitive();
 }
 
 void gui_clear_digests(void)
 {
-	switch (gui_get_view()) {
+	switch (gui.view) {
 		case GUI_VIEW_FILE:
-			for (int i = 0; i < HASH_FUNCS_N; i++)
+			for (int i = 0; i < HASH_FUNCS_N; i++) {
+				if (!hash.funcs[i].supported)
+					continue;
 				gtk_entry_set_text(gui.hash_widgets[i].entry_file, "");
+			}
 			gui_check_digests();
 			break;
 		case GUI_VIEW_TEXT:
-			for (int i = 0; i < HASH_FUNCS_N; i++)
+			for (int i = 0; i < HASH_FUNCS_N; i++) {
+				if (!hash.funcs[i].supported)
+					continue;
 				gtk_entry_set_text(gui.hash_widgets[i].entry_text, "");
+			}
 			gui_check_digests();
 			break;
 		case GUI_VIEW_FILE_LIST:
@@ -665,88 +794,92 @@ void gui_clear_digests(void)
 		default:
 			g_assert_not_reached();
 	}
-}
 
-void gui_clear_all_digests(void)
-{
-	for (int i = 0; i < HASH_FUNCS_N; i++)
-		gtk_entry_set_text(gui.hash_widgets[i].entry_file, "");
-
-	for (int i = 0; i < HASH_FUNCS_N; i++)
-		gtk_entry_set_text(gui.hash_widgets[i].entry_text, "");
-
-	list_clear_digests();
-
-	gui_check_digests();
+	gui_menuitem_save_as_set_sensitive();
 }
 
 void gui_check_digests(void)
 {
-	const char *str_in_file = gtk_entry_get_text(gui.entry_check_file);
-	const char *str_in_text = gtk_entry_get_text(gui.entry_check_text);
+	const char *icon_in = NULL;
+	GtkEntry *entry_check = NULL;
 
-	const char *icon_in_file = NULL;
-	const char *icon_in_text = NULL;
+	switch (gui.view) {
+		case GUI_VIEW_TEXT:
+			entry_check = gui.entry_check_text;
+			break;
+		case GUI_VIEW_FILE:
+			entry_check = gui.entry_check_file;
+			break;
+		case GUI_VIEW_FILE_LIST: // TODO
+			return;
+		default:
+			g_assert_not_reached();
+	}
+
+	const enum digest_format_e format = gui_get_digest_format();
+	const char *str_in = gtk_entry_get_text(entry_check);
 
 	for (int i = 0; i < HASH_FUNCS_N; i++) {
 		if (!hash.funcs[i].enabled)
 			continue;
 
-		const char *str_out_file = gtk_entry_get_text(
-			gui.hash_widgets[i].entry_file);
-		const char *str_out_text = gtk_entry_get_text(
-			gui.hash_widgets[i].entry_text);
+		GtkEntry *entry = NULL;
 
-		const char *icon_out_file = NULL;
-		const char *icon_out_text = NULL;
-
-		switch (gui_get_digest_format()) {
-			case DIGEST_FORMAT_HEX_LOWER:
-			case DIGEST_FORMAT_HEX_UPPER:
-				if (*str_in_file && (strcasecmp(str_in_file, str_out_file) == 0))
-					icon_out_file = GTK_STOCK_YES;
-				if (*str_in_text && (strcasecmp(str_in_text, str_out_text) == 0))
-					icon_out_text = GTK_STOCK_YES;
+		switch (gui.view) {
+			case GUI_VIEW_TEXT:
+				entry = gui.hash_widgets[i].entry_text;
 				break;
-			case DIGEST_FORMAT_BASE64:
-				if (*str_in_file && (strcmp(str_in_file, str_out_file) == 0))
-					icon_out_file = GTK_STOCK_YES;
-				if (*str_in_text && (strcmp(str_in_text, str_out_text) == 0))
-					icon_out_text = GTK_STOCK_YES;
+			case GUI_VIEW_FILE:
+				entry = gui.hash_widgets[i].entry_file;
 				break;
 			default:
 				g_assert_not_reached();
 		}
 
-		gtk_entry_set_icon_from_stock(gui.hash_widgets[i].entry_file,
-			GTK_ENTRY_ICON_SECONDARY, icon_out_file);
-		gtk_entry_set_icon_from_stock(gui.hash_widgets[i].entry_text,
-			GTK_ENTRY_ICON_SECONDARY, icon_out_text);
+		const char *str_out = gtk_entry_get_text(entry);
+		const char *icon_out = NULL;
 
-		if (icon_out_file)
-			icon_in_file = GTK_STOCK_YES;
-		if (icon_out_text)
-			icon_in_text = GTK_STOCK_YES;
+		switch (format) {
+			case DIGEST_FORMAT_HEX_LOWER:
+			case DIGEST_FORMAT_HEX_UPPER:
+				if (*str_in && (g_ascii_strcasecmp(str_in, str_out) == 0)) {
+					// FIXME: find a real alternative for GTK_STOCK_YES
+					icon_out = "gtk-yes";
+					icon_in = "gtk-yes";
+				}
+				break;
+			case DIGEST_FORMAT_BASE64:
+				if (*str_in && (strcmp(str_in, str_out) == 0)) {
+					icon_out = "gtk-yes";
+					icon_in = "gtk-yes";
+				}
+				break;
+			default:
+				g_assert_not_reached();
+		}
+
+		gtk_entry_set_icon_from_icon_name(entry, GTK_ENTRY_ICON_SECONDARY,
+			icon_out);
 	}
 
-	gtk_entry_set_icon_from_stock(gui.entry_check_file,
-		GTK_ENTRY_ICON_SECONDARY, icon_in_file);
-	gtk_entry_set_icon_from_stock(gui.entry_check_text,
-		GTK_ENTRY_ICON_SECONDARY, icon_in_text);
+	gtk_entry_set_icon_from_icon_name(entry_check, GTK_ENTRY_ICON_SECONDARY,
+		icon_in);
 }
 
 void gui_set_state(const enum gui_state_e state)
 {
 	g_assert(GUI_STATE_IS_VALID(state));
+	g_assert(state != gui_priv.state);
+
+	if (gui.view == GUI_VIEW_TEXT)
+		g_assert(state != GUI_STATE_BUSY);
 
 	gui_priv.state = state;
 
-	if (gui_get_view() == GUI_VIEW_TEXT)
-		return;
-
 	const bool busy = (state == GUI_STATE_BUSY);
 
 	gtk_widget_set_visible(GTK_WIDGET(gui.button_hash), !busy);
+
 	gtk_widget_set_visible(GTK_WIDGET(gui.button_stop), busy);
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.button_stop), busy);
 
@@ -759,28 +892,25 @@ void gui_set_state(const enum gui_state_e state)
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.toolbar), !busy);
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.treeview), !busy);
 
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_open), !busy);
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.radiomenuitem_text), !busy);
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.radiomenuitem_file), !busy);
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.radiomenuitem_file_list), !busy);
 
+#if (GTK_MAJOR_VERSION > 2)
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.dialog_grid), !busy);
+#else
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.dialog_table), !busy);
+#endif
+
 	gtk_widget_set_sensitive(GTK_WIDGET(gui.dialog_combobox), !busy);
 
-	if (busy) {
+	if (busy)
 		gtk_window_set_default(gui.window, GTK_WIDGET(gui.button_stop));
-		gtk_widget_set_sensitive(GTK_WIDGET(gui.menuitem_save_as), false);
-	} else {
+	else
 		gtk_window_set_default(gui.window, GTK_WIDGET(gui.button_hash));
-		// User may already have menu open, so make sure save_as gets updated
-		gtk_menu_item_activate(gui.menuitem_file);
-	}
-}
 
-enum gui_state_e gui_get_state(void)
-{
-	g_assert(GUI_STATE_IS_VALID(gui_priv.state));
-
-	return gui_priv.state;
+	gui_menuitem_save_as_set_sensitive();
 }
 
 bool gui_is_maximised(void)
@@ -794,3 +924,32 @@ bool gui_is_maximised(void)
 
 	return (state & GDK_WINDOW_STATE_MAXIMIZED);
 }
+
+void gui_start_hash(void)
+{
+	switch (gui.view) {
+		case GUI_VIEW_FILE: {
+			gui_clear_digests();
+			gui_set_state(GUI_STATE_BUSY);
+			char *uri = gtk_file_chooser_get_uri(GTK_FILE_CHOOSER(
+				gui.filechooserbutton));
+			hash_file_start(uri);
+			break;
+		}
+		case GUI_VIEW_FILE_LIST:
+			gui_clear_digests();
+			gui_set_state(GUI_STATE_BUSY);
+			hash_file_list_start();
+			break;
+		default:
+			g_assert_not_reached();
+	}
+}
+
+void gui_stop_hash(void)
+{
+	g_assert(gui.view != GUI_VIEW_TEXT);
+
+	gtk_widget_set_sensitive(GTK_WIDGET(gui.button_stop), false);
+	hash_file_stop();
+}
