@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2007-2011 Tristan Heaven <tristanheaven@gmail.com>
+ *   Copyright (C) 2007-2016 Tristan Heaven <tristan@tristanheaven.net>
  *
  *   This file is part of GtkHash.
  *
@@ -14,7 +14,7 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with GtkHash. If not, see <http://www.gnu.org/licenses/>.
+ *   along with GtkHash. If not, see <https://gnu.org/licenses/gpl-2.0.txt>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -29,6 +29,12 @@
 #if IN_NAUTILUS_EXTENSION
 	#include <libnautilus-extension/nautilus-property-page.h>
 	#include <libnautilus-extension/nautilus-property-page-provider.h>
+#elif IN_CAJA_EXTENSION
+	#include <libcaja-extension/caja-property-page.h>
+	#include <libcaja-extension/caja-property-page-provider.h>
+#elif IN_NEMO_EXTENSION
+	#include <libnemo-extension/nemo-property-page.h>
+	#include <libnemo-extension/nemo-property-page-provider.h>
 #elif IN_THUNAR_EXTENSION
 	#include <thunarx/thunarx.h>
 #endif
@@ -37,10 +43,11 @@
 #include "properties-list.h"
 #include "properties-hash.h"
 #include "properties-prefs.h"
+#include "properties-resources.h"
 #include "../hash/hash-func.h"
 #include "../hash/hash-file.h"
 
-#define BUILDER_XML DATADIR "/gtkhash-properties.xml.gz"
+#define PROPERTIES_XML_RESOURCE "/org/gtkhash/plugin/gtkhash-properties.xml"
 
 static GType page_type;
 
@@ -58,8 +65,6 @@ static void gtkhash_properties_busy(struct page_s *page)
 {
 	page->busy = true;
 
-	gtk_widget_set_sensitive(GTK_WIDGET(page->button_hash), false);
-	gtk_widget_set_sensitive(GTK_WIDGET(page->button_stop), true);
 	gtk_widget_set_sensitive(GTK_WIDGET(page->treeview), false);
 	gtk_widget_set_sensitive(GTK_WIDGET(page->hbox_inputs), false);
 
@@ -67,6 +72,14 @@ static void gtkhash_properties_busy(struct page_s *page)
 	gtk_progress_bar_set_fraction(page->progressbar, 0.0);
 	gtk_progress_bar_set_text(page->progressbar, " ");
 	gtk_widget_show(GTK_WIDGET(page->progressbar));
+
+	// Hash button
+	gtk_widget_hide(GTK_WIDGET(page->button_hash));
+	gtk_widget_set_sensitive(GTK_WIDGET(page->button_hash), false);
+
+	// Stop button
+	gtk_widget_set_sensitive(GTK_WIDGET(page->button_stop), true);
+	gtk_widget_show(GTK_WIDGET(page->button_stop));
 }
 
 static void gtkhash_properties_button_hash_set_sensitive(struct page_s *page)
@@ -74,7 +87,7 @@ static void gtkhash_properties_button_hash_set_sensitive(struct page_s *page)
 	bool has_enabled = false;
 
 	for (int i = 0; i < HASH_FUNCS_N; i++) {
-		if (page->hash_file.funcs[i].enabled) {
+		if (page->funcs[i].enabled) {
 			has_enabled = true;
 			break;
 		}
@@ -96,11 +109,17 @@ void gtkhash_properties_idle(struct page_s *page)
 
 	gtk_widget_hide(GTK_WIDGET(page->progressbar));
 
+	// Stop button
+	gtk_widget_hide(GTK_WIDGET(page->button_stop));
 	gtk_widget_set_sensitive(GTK_WIDGET(page->button_stop), false);
+
+	// Hash button
+	gtkhash_properties_button_hash_set_sensitive(page);
+	gtk_widget_show(GTK_WIDGET(page->button_hash));
+
 	gtk_widget_set_sensitive(GTK_WIDGET(page->treeview), true);
 	gtk_widget_set_sensitive(GTK_WIDGET(page->hbox_inputs), true);
 	gtkhash_properties_entry_hmac_set_sensitive(page);
-	gtkhash_properties_button_hash_set_sensitive(page);
 
 	gtkhash_properties_list_check_digests(page);
 }
@@ -131,7 +150,7 @@ static bool gtkhash_properties_on_treeview_button_press_event(
 	return false;
 }
 
-static void gtkhash_properties_on_treeselection_changed(struct page_s *page)
+static void gtkhash_properties_on_menu_map_event(struct page_s *page)
 {
 	bool sensitive = false;
 	char *digest = gtkhash_properties_list_get_selected_digest(page);
@@ -166,27 +185,55 @@ static void gtkhash_properties_on_entry_check_changed(struct page_s *page)
 
 static void gtkhash_properties_on_entry_hmac_changed(struct page_s *page)
 {
-	gtkhash_hash_file_clear_digests(&page->hash_file);
-	gtkhash_properties_list_update_digests(page);
+	gtkhash_properties_list_clear_digests(page);
 	gtkhash_properties_list_check_digests(page);
 }
 
+static void gtkhash_properties_on_menuitem_show_hmac_key_toggled(
+	GtkCheckMenuItem *item, GtkEntry *entry)
+{
+	const bool active = gtk_check_menu_item_get_active(item);
+
+	gtk_entry_set_visibility(entry, active);
+}
+
+static void gtkhash_properties_on_entry_hmac_populate_popup(GtkEntry *entry,
+	GtkMenu *menu)
+{
+	GtkWidget *item;
+
+	// Add separator
+	item = gtk_separator_menu_item_new();
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
+	// Add checkbutton
+	item = gtk_check_menu_item_new_with_mnemonic(_("_Show HMAC Key"));
+	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),
+		gtk_entry_get_visibility(entry));
+	gtk_widget_show(item);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+	g_signal_connect(item, "toggled", G_CALLBACK(
+		gtkhash_properties_on_menuitem_show_hmac_key_toggled), entry);
+}
+
 static void gtkhash_properties_on_togglebutton_hmac_toggled(struct page_s *page)
 {
 	gtkhash_properties_entry_hmac_set_sensitive(page);
-	gtkhash_properties_on_entry_hmac_changed(page);
+	gtkhash_properties_list_update_hash_func_names(page);
+	gtkhash_properties_list_check_digests(page);
 }
 
 static void gtkhash_properties_on_button_hash_clicked(struct page_s *page)
 {
 	gtkhash_properties_busy(page);
-	gtkhash_hash_file_clear_digests(&page->hash_file);
-	gtkhash_properties_list_update_digests(page);
+	gtkhash_properties_list_clear_digests(page);
 
 	if (gtk_toggle_button_get_active(page->togglebutton_hmac)) {
 		const uint8_t *hmac_key = (uint8_t *)gtk_entry_get_text(
 			page->entry_hmac);
-		const size_t key_size = gtk_entry_get_text_length(page->entry_hmac);
+		GtkEntryBuffer *buffer = gtk_entry_get_buffer(page->entry_hmac);
+		const size_t key_size = gtk_entry_buffer_get_bytes(buffer);
 		gtkhash_properties_hash_start(page, hmac_key, key_size);
 	} else
 		gtkhash_properties_hash_start(page, NULL, 0);
@@ -213,12 +260,13 @@ static void gtkhash_properties_free_page(struct page_s *page)
 	g_free(page);
 }
 
-static void gtkhash_properties_get_objects(struct page_s *page,
+static void gtkhash_properties_init_objects(struct page_s *page,
 	GtkBuilder *builder)
 {
 	// Main container
 	page->box = GTK_WIDGET(gtkhash_properties_get_object(builder,
 		"vbox"));
+	g_object_ref(page->box);
 
 	// Progress bar
 	page->progressbar = GTK_PROGRESS_BAR(gtkhash_properties_get_object(builder,
@@ -235,7 +283,8 @@ static void gtkhash_properties_get_objects(struct page_s *page,
 	// Popup menu
 	page->menu = GTK_MENU(gtkhash_properties_get_object(builder,
 		"menu"));
-	page->menuitem_copy = GTK_IMAGE_MENU_ITEM(gtkhash_properties_get_object(builder,
+	g_object_ref(page->menu);
+	page->menuitem_copy = GTK_MENU_ITEM(gtkhash_properties_get_object(builder,
 		"imagemenuitem_copy"));
 	page->menuitem_show_funcs = GTK_CHECK_MENU_ITEM(gtkhash_properties_get_object(builder,
 		"checkmenuitem_show_funcs"));
@@ -255,6 +304,10 @@ static void gtkhash_properties_get_objects(struct page_s *page,
 		"button_hash"));
 	page->button_stop = GTK_BUTTON(gtkhash_properties_get_object(builder,
 		"button_stop"));
+#if (GTK_MAJOR_VERSION > 2)
+	gtk_button_set_always_show_image(page->button_hash, true);
+	gtk_button_set_always_show_image(page->button_stop, true);
+#endif
 }
 
 static void gtkhash_properties_connect_signals(struct page_s *page)
@@ -270,10 +323,10 @@ static void gtkhash_properties_connect_signals(struct page_s *page)
 		G_CALLBACK(gtkhash_properties_on_treeview_popup_menu), page);
 	g_signal_connect_swapped(page->treeview, "button-press-event",
 		G_CALLBACK(gtkhash_properties_on_treeview_button_press_event), page);
-	g_signal_connect_swapped(page->treeselection, "changed",
-		G_CALLBACK(gtkhash_properties_on_treeselection_changed), page);
 
 	// Popup menu
+	g_signal_connect_swapped(page->menu, "map-event",
+		G_CALLBACK(gtkhash_properties_on_menu_map_event), page);
 	g_signal_connect_swapped(page->menuitem_copy, "activate",
 		G_CALLBACK(gtkhash_properties_on_menuitem_copy_activate), page);
 	g_signal_connect_swapped(page->menuitem_show_funcs, "toggled",
@@ -288,6 +341,8 @@ static void gtkhash_properties_connect_signals(struct page_s *page)
 		G_CALLBACK(gtkhash_properties_on_togglebutton_hmac_toggled), page);
 	g_signal_connect_swapped(page->entry_hmac, "changed",
 		G_CALLBACK(gtkhash_properties_on_entry_hmac_changed), page);
+	g_signal_connect(page->entry_hmac, "populate-popup",
+		G_CALLBACK(gtkhash_properties_on_entry_hmac_populate_popup), NULL);
 
 	// Buttons
 	g_signal_connect_swapped(page->button_hash, "clicked",
@@ -296,73 +351,46 @@ static void gtkhash_properties_connect_signals(struct page_s *page)
 		G_CALLBACK(gtkhash_properties_on_button_stop_clicked), page);
 }
 
-static char *gtkhash_properties_get_xml(const char *filename)
+static GtkBuilder *gtkhash_properties_init_builder(void)
 {
-	GMappedFile *map = g_mapped_file_new(filename, false, NULL);
+#if (GTK_MAJOR_VERSION > 2)
+	return gtk_builder_new_from_resource(PROPERTIES_XML_RESOURCE);
+#else
+	GError *error = NULL;
+	GBytes *bytes = g_resources_lookup_data(PROPERTIES_XML_RESOURCE,
+		G_RESOURCE_LOOKUP_FLAGS_NONE, &error);
 
-	if (!map)
-		return NULL;
+	if (G_UNLIKELY(error)) {
+		g_warning("%s", error->message);
+		g_error_free(error);
 
-	gsize map_len = g_mapped_file_get_length(map);
-	if (map_len == 0) {
-		g_mapped_file_unref(map);
 		return NULL;
 	}
 
-	const char *map_data = g_mapped_file_get_contents(map);
-	g_assert(map_data);
+	gsize xml_len = 0;
+	char *xml = g_bytes_unref_to_data(bytes, &xml_len);
+	GtkBuilder *builder = gtk_builder_new();
 
-	GInputStream *input_mem = g_memory_input_stream_new_from_data(map_data,
-		map_len, NULL);
-	GConverter *converter = G_CONVERTER(g_zlib_decompressor_new(
-		G_ZLIB_COMPRESSOR_FORMAT_GZIP));
+	gtk_builder_add_from_string(builder, xml, xml_len, &error);
+	g_free(xml);
 
-	GInputStream *input_conv = g_converter_input_stream_new(input_mem,
-		converter);
+	if (G_UNLIKELY(error)) {
+		g_warning("%s", error->message);
+		g_error_free(error);
+		g_object_unref(builder);
 
-	g_object_unref(input_mem);
-	g_object_unref(converter);
-
-	GString *string = g_string_new(NULL);
-
-	for (char buf[1024];;) {
-		gssize len = g_input_stream_read(input_conv, buf, 1024, NULL, NULL);
-		if (len <= 0)
-			break;
-
-		g_string_append_len(string, buf, len);
+		return NULL;
 	}
 
-	g_object_unref(input_conv);
-	g_mapped_file_unref(map);
-
-	return g_string_free(string, false);
+	return builder;
+#endif
 }
 
 static struct page_s *gtkhash_properties_new_page(char *uri)
 {
-	char *xml = gtkhash_properties_get_xml(BUILDER_XML);
-
-	if (!xml || !*xml) {
-		g_warning("failed to read \"%s\"", BUILDER_XML);
-		g_free(xml);
+	GtkBuilder *builder = gtkhash_properties_init_builder();
+	if (!builder)
 		return NULL;
-	}
-
-	GtkBuilder *builder = gtk_builder_new();
-	gtk_builder_set_translation_domain(builder, GETTEXT_PACKAGE);
-
-	GError *error = NULL;
-	gtk_builder_add_from_string(builder, xml, -1, &error);
-
-	g_free(xml);
-
-	if (error) {
-		g_warning("failed to read \"%s\":\n%s", BUILDER_XML, error->message);
-		g_error_free(error);
-		g_object_unref(builder);
-		return NULL;
-	}
 
 	struct page_s *page = g_new(struct page_s, 1);
 	page->uri = uri;
@@ -373,12 +401,11 @@ static struct page_s *gtkhash_properties_new_page(char *uri)
 		g_warning("no hash functions available");
 		gtkhash_properties_hash_deinit(page);
 		g_free(page);
+
 		return NULL;
 	}
 
-	gtkhash_properties_get_objects(page, builder);
-	g_object_ref(page->box);
-	g_object_ref(page->menu);
+	gtkhash_properties_init_objects(page, builder);
 	g_object_unref(builder);
 
 	gtkhash_properties_prefs_init(page);
@@ -393,6 +420,10 @@ static struct page_s *gtkhash_properties_new_page(char *uri)
 static GList *gtkhash_properties_get_pages(
 #if IN_NAUTILUS_EXTENSION
 	G_GNUC_UNUSED NautilusPropertyPageProvider *provider,
+#elif IN_CAJA_EXTENSION
+	G_GNUC_UNUSED CajaPropertyPageProvider *provider,
+#elif IN_NEMO_EXTENSION
+	G_GNUC_UNUSED NemoPropertyPageProvider *provider,
 #elif IN_THUNAR_EXTENSION
 	G_GNUC_UNUSED ThunarxPropertyPageProvider *provider,
 #endif
@@ -404,8 +435,13 @@ static GList *gtkhash_properties_get_pages(
 
 #if IN_NAUTILUS_EXTENSION
 	GFileType type = nautilus_file_info_get_file_type(files->data);
-
 	char *uri = nautilus_file_info_get_uri(files->data);
+#elif IN_CAJA_EXTENSION
+	GFileType type = caja_file_info_get_file_type(files->data);
+	char *uri = caja_file_info_get_uri(files->data);
+#elif IN_NEMO_EXTENSION
+	GFileType type = nemo_file_info_get_file_type(files->data);
+	char *uri = nemo_file_info_get_uri(files->data);
 #elif IN_THUNAR_EXTENSION
 	GFileInfo *info = thunarx_file_info_get_file_info(files->data);
 	GFileType type = g_file_info_get_file_type(info);
@@ -425,6 +461,12 @@ static GList *gtkhash_properties_get_pages(
 #if IN_NAUTILUS_EXTENSION
 	NautilusPropertyPage *ppage = nautilus_property_page_new(
 		"GtkHash::properties", gtk_label_new(_("Digests")), page->box);
+#elif IN_CAJA_EXTENSION
+	CajaPropertyPage *ppage = caja_property_page_new(
+		"GtkHash::properties", gtk_label_new(_("Digests")), page->box);
+#elif IN_NEMO_EXTENSION
+	NemoPropertyPage *ppage = nemo_property_page_new(
+		"GtkHash::properties", gtk_label_new(_("Digests")), page->box);
 #elif IN_THUNAR_EXTENSION
 	GtkWidget *ppage = thunarx_property_page_new(_("Digests"));
 	gtk_container_add(GTK_CONTAINER(ppage), page->box);
@@ -438,6 +480,10 @@ static GList *gtkhash_properties_get_pages(
 static void gtkhash_properties_iface_init(
 #if IN_NAUTILUS_EXTENSION
 	NautilusPropertyPageProviderIface *iface
+#elif IN_CAJA_EXTENSION
+	CajaPropertyPageProviderIface *iface
+#elif IN_NEMO_EXTENSION
+	NemoPropertyPageProviderIface *iface
 #elif IN_THUNAR_EXTENSION
 	ThunarxPropertyPageProviderIface *iface
 #endif
@@ -473,6 +519,10 @@ static void gtkhash_properties_register_type(GTypeModule *module)
 	g_type_module_add_interface(module, page_type,
 #if IN_NAUTILUS_EXTENSION
 		NAUTILUS_TYPE_PROPERTY_PAGE_PROVIDER,
+#elif IN_CAJA_EXTENSION
+		CAJA_TYPE_PROPERTY_PAGE_PROVIDER,
+#elif IN_NEMO_EXTENSION
+		NEMO_TYPE_PROPERTY_PAGE_PROVIDER,
 #elif IN_THUNAR_EXTENSION
 		THUNARX_TYPE_PROPERTY_PAGE_PROVIDER,
 #endif
@@ -487,21 +537,30 @@ static void gtkhash_properties_register_type(GTypeModule *module)
 
 #if IN_NAUTILUS_EXTENSION
 PUBLIC void nautilus_module_initialize(GTypeModule *module)
+#elif IN_CAJA_EXTENSION
+PUBLIC void caja_module_initialize(GTypeModule *module)
+#elif IN_NEMO_EXTENSION
+PUBLIC void nemo_module_initialize(GTypeModule *module)
 #elif IN_THUNAR_EXTENSION
 PUBLIC void thunar_extension_initialize(GTypeModule *module);
 PUBLIC void thunar_extension_initialize(GTypeModule *module)
 #endif
 {
+	gtkhash_properties_resources_register_resource();
 	gtkhash_properties_register_type(module);
 
 #if ENABLE_NLS
 	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
-	bind_textdomain_codeset(PACKAGE, "UTF-8");
+	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
 #endif
 }
 
 #if IN_NAUTILUS_EXTENSION
 PUBLIC void nautilus_module_shutdown(void)
+#elif IN_CAJA_EXTENSION
+PUBLIC void caja_module_shutdown(void)
+#elif IN_NEMO_EXTENSION
+PUBLIC void nemo_module_shutdown(void)
 #elif IN_THUNAR_EXTENSION
 PUBLIC void thunar_extension_shutdown(void);
 PUBLIC void thunar_extension_shutdown(void)
@@ -511,6 +570,10 @@ PUBLIC void thunar_extension_shutdown(void)
 
 #if IN_NAUTILUS_EXTENSION
 PUBLIC void nautilus_module_list_types(const GType **types, int *num_types)
+#elif IN_CAJA_EXTENSION
+PUBLIC void caja_module_list_types(const GType **types, int *num_types)
+#elif IN_NEMO_EXTENSION
+PUBLIC void nemo_module_list_types(const GType **types, int *num_types)
 #elif IN_THUNAR_EXTENSION
 PUBLIC void thunar_extension_list_types(const GType **types, int *num_types);
 PUBLIC void thunar_extension_list_types(const GType **types, int *num_types)
