@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2007-2011 Tristan Heaven <tristanheaven@gmail.com>
+ *   Copyright (C) 2007-2016 Tristan Heaven <tristan@tristanheaven.net>
  *
  *   This file is part of GtkHash.
  *
@@ -14,7 +14,7 @@
  *   GNU General Public License for more details.
  *
  *   You should have received a copy of the GNU General Public License
- *   along with GtkHash. If not, see <http://www.gnu.org/licenses/>.
+ *   along with GtkHash. If not, see <https://gnu.org/licenses/gpl-2.0.txt>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -22,6 +22,7 @@
 #endif
 
 #include <stdlib.h>
+#include <string.h>
 #include <stdbool.h>
 #include <gtk/gtk.h>
 
@@ -32,7 +33,7 @@
 #include "hash/digest-format.h"
 #include "hash/hash-func.h"
 
-#define PREFS_SCHEMA "apps.gtkhash"
+#define PREFS_SCHEMA "org.gtkhash"
 #define PREFS_KEY_DIGEST_FORMAT "digest-format"
 #define PREFS_KEY_HASH_FUNCS "hash-functions"
 #define PREFS_KEY_SHOW_TOOLBAR "show-toolbar"
@@ -52,6 +53,10 @@ static struct {
 
 static void default_hash_funcs(void)
 {
+	// Some funcs could already be enabled by cmdline arg
+	if (hash_funcs_count_enabled())
+		return;
+
 	bool has_enabled = false;
 
 	// Try to enable default functions
@@ -87,27 +92,18 @@ static void prefs_default(void)
 {
 	default_hash_funcs();
 	default_show_toolbar();
-
-	gui_update();
 }
 
 static void load_hash_funcs(void)
 {
-	char **strv = g_settings_get_strv(prefs_priv.settings, PREFS_KEY_HASH_FUNCS);
-	bool has_enabled = false;
+	char **strv = g_settings_get_strv(prefs_priv.settings,
+		PREFS_KEY_HASH_FUNCS);
 
-	for (int i = 0; strv[i]; i++) {
-		enum hash_func_e id = gtkhash_hash_func_get_id_from_name(strv[i]);
-		if (id != HASH_FUNC_INVALID && hash.funcs[id].supported) {
-			hash.funcs[id].enabled = true;
-			gtk_toggle_button_set_active(gui.hash_widgets[id].button, true);
-			has_enabled = true;
-		}
-	}
+	hash_funcs_enable_strv((const char **)strv);
 
 	g_strfreev(strv);
 
-	if (!has_enabled)
+	if (!hash_funcs_count_enabled())
 		default_hash_funcs();
 }
 
@@ -120,11 +116,11 @@ static void load_digest_format(void)
 
 	enum digest_format_e format = DIGEST_FORMAT_INVALID;
 
-	if (g_strcmp0(str, "hex-lower") == 0)
+	if (strcmp(str, "hex-lower") == 0)
 		format = DIGEST_FORMAT_HEX_LOWER;
-	else if (g_strcmp0(str, "hex-upper") == 0)
+	else if (strcmp(str, "hex-upper") == 0)
 		format = DIGEST_FORMAT_HEX_UPPER;
-	else if (g_strcmp0(str, "base64") == 0)
+	else if (strcmp(str, "base64") == 0)
 		format = DIGEST_FORMAT_BASE64;
 
 	g_free(str);
@@ -139,19 +135,14 @@ static void load_view(void)
 	if (!str)
 		return;
 
-	enum gui_view_e view = GUI_VIEW_INVALID;
-
-	if (g_strcmp0(str, "file") == 0)
-		view = GUI_VIEW_FILE;
-	else if (g_strcmp0(str, "text") == 0)
-		view = GUI_VIEW_TEXT;
-	else if (g_strcmp0(str, "file-list") == 0)
-		view = GUI_VIEW_FILE_LIST;
+	if (strcmp(str, "file") == 0)
+		gui_set_view(GUI_VIEW_FILE);
+	else if (strcmp(str, "text") == 0)
+		gui_set_view(GUI_VIEW_TEXT);
+	else if (strcmp(str, "file-list") == 0)
+		gui_set_view(GUI_VIEW_FILE_LIST);
 
 	g_free(str);
-
-	if (GUI_VIEW_IS_VALID(view))
-		gui_set_view(view);
 }
 
 static void load_show_toolbar(void)
@@ -183,20 +174,12 @@ static void prefs_load(void)
 	load_view();
 	load_show_toolbar();
 	load_window_size();
-
-	gui_update();
 }
 
 static void save_hash_funcs(void)
 {
 	const char **strv = NULL;
-	int enabled = 0;
-
-	for (int i = 0; i < HASH_FUNCS_N; i++) {
-		if (hash.funcs[i].enabled) {
-			enabled++;
-		}
-	}
+	int enabled = hash_funcs_count_enabled();
 
 	if (enabled > 0) {
 		strv = g_new0(const char *, enabled + 1);
@@ -239,7 +222,7 @@ static void save_view(void)
 {
 	const char *str = NULL;
 
-	switch (gui_get_view()) {
+	switch (gui.view) {
 		case GUI_VIEW_FILE:
 			str = "file";
 			break;
@@ -284,17 +267,11 @@ void prefs_init(void)
 {
 	g_assert(!prefs_priv.settings);
 
-	bool found = false;
-	const char * const *schemas = g_settings_list_schemas();
+	GSettingsSchema *schema = g_settings_schema_source_lookup(
+		g_settings_schema_source_get_default(), PREFS_SCHEMA, true);
 
-	for (int i = 0; schemas[i]; i++) {
-		if (g_strcmp0(schemas[i], PREFS_SCHEMA) == 0) {
-			found = true;
-			break;
-		}
-	}
-
-	if (found) {
+	if (schema) {
+		g_settings_schema_unref(schema);
 		prefs_priv.settings = g_settings_new(PREFS_SCHEMA);
 		prefs_load();
 	} else {
